import PropTypes from 'prop-types';
import React, { createContext, useState, useRef, useCallback, useMemo, useContext, useEffect, useLayoutEffect } from 'react';

let StatusEnum;

(function (StatusEnum) {
  StatusEnum["ERROR"] = "error";
  StatusEnum["IDLE"] = "idle";
  StatusEnum["LOADING"] = "loading";
  StatusEnum["SUCCESS"] = "success";
})(StatusEnum || (StatusEnum = {}));

let ActionEnum;

(function (ActionEnum) {
  ActionEnum["ON_ERROR"] = "ON_ERROR";
  ActionEnum["ON_LOADING"] = "ON_LOADING";
  ActionEnum["ON_SUCCESS"] = "ON_SUCCESS";
  ActionEnum["ON_UPDATE_DATA"] = "ON_UPDATE_DATA";
  ActionEnum["RESET"] = "RESET";
})(ActionEnum || (ActionEnum = {}));

const KEY_IS_NOT_STRING_ERROR = 'Key should be a string';

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

class DataLoader {
  constructor(args) {
    _defineProperty(this, "key", void 0);

    _defineProperty(this, "status", void 0);

    _defineProperty(this, "pollingInterval", void 0);

    _defineProperty(this, "maxDataLifetime", void 0);

    _defineProperty(this, "isDataOutdated", false);

    _defineProperty(this, "notify", void 0);

    _defineProperty(this, "method", void 0);

    _defineProperty(this, "cancelMethod", void 0);

    _defineProperty(this, "canceled", false);

    _defineProperty(this, "keepPreviousData", void 0);

    _defineProperty(this, "errorListeners", []);

    _defineProperty(this, "successListeners", []);

    _defineProperty(this, "cancelListeners", []);

    _defineProperty(this, "error", void 0);

    _defineProperty(this, "dataOutdatedTimeout", void 0);

    _defineProperty(this, "timeout", void 0);

    _defineProperty(this, "load", async (force = false) => {
      if (force || this.status !== StatusEnum.SUCCESS || this.status === StatusEnum.SUCCESS && this.isDataOutdated) {
        if (this.timeout) {
          // Prevent multiple call at the same time
          clearTimeout(this.timeout);
        }

        await this.launch();
      }
    });

    _defineProperty(this, "launch", async () => {
      try {
        if (this.status !== StatusEnum.LOADING) {
          this.canceled = false;
          this.status = StatusEnum.LOADING;
          this.notify(this);
        }

        const promise = this.method();
        this.cancelMethod = promise.cancel;
        const result = await promise.then(res => res);
        this.status = StatusEnum.SUCCESS;
        this.error = undefined;
        this.notify(this);

        if (!this.canceled) {
          await Promise.all(this.successListeners.map(listener => listener === null || listener === void 0 ? void 0 : listener(result)));
          this.isDataOutdated = false;

          if (this.dataOutdatedTimeout) {
            clearTimeout(this.dataOutdatedTimeout);
            this.dataOutdatedTimeout = undefined;
          }

          if (this.maxDataLifetime) {
            this.dataOutdatedTimeout = setTimeout(() => {
              this.isDataOutdated = true;
            }, this.maxDataLifetime);
          }
        }
      } catch (err) {
        this.status = StatusEnum.ERROR;
        this.error = err;
        this.notify(this);

        if (!this.canceled) {
          await Promise.all(this.errorListeners.map(listener => listener === null || listener === void 0 ? void 0 : listener(err)));
        }
      }

      if (this.pollingInterval) {
        this.timeout = setTimeout( // eslint-disable-next-line @typescript-eslint/no-misused-promises
        this.launch, this.pollingInterval);
      }
    });

    _defineProperty(this, "cancel", async () => {
      var _this$cancelMethod;

      this.canceled = true;
      (_this$cancelMethod = this.cancelMethod) === null || _this$cancelMethod === void 0 ? void 0 : _this$cancelMethod.call(this);
      await Promise.all(this.cancelListeners.map(listener => listener === null || listener === void 0 ? void 0 : listener()));
    });

    this.key = args.key;
    this.status = args.enabled ? StatusEnum.LOADING : StatusEnum.IDLE;
    this.method = args.method;
    this.pollingInterval = args === null || args === void 0 ? void 0 : args.pollingInterval;
    this.keepPreviousData = args === null || args === void 0 ? void 0 : args.keepPreviousData;
    this.notify = args.notify;
    this.maxDataLifetime = args.maxDataLifetime;
  }

  addOnSuccessListener(fn) {
    if (!this.successListeners.includes(fn)) {
      this.successListeners.push(fn);
    }
  }

  addOnErrorListener(fn) {
    if (!this.errorListeners.includes(fn)) {
      this.errorListeners.push(fn);
    }
  }

  addOnCancelListener(fn) {
    if (!this.cancelListeners.includes(fn)) {
      this.cancelListeners.push(fn);
    }
  }

  removeOnSuccessListener(fn) {
    const index = this.successListeners.indexOf(fn);

    if (index > -1) {
      this.successListeners.splice(index, 1);
    }
  }

  removeOnErrorListener(fn) {
    const index = this.errorListeners.indexOf(fn);

    if (index > -1) {
      this.errorListeners.splice(index, 1);
    }
  }

  removeOnCancelListener(fn) {
    const index = this.cancelListeners.indexOf(fn);

    if (index > -1) {
      this.cancelListeners.splice(index, 1);
    }
  }

  destroy() {
    if (this.timeout) {
      clearTimeout(this.timeout);
    }
  }

}

// @ts-expect-error we force the context to undefined, should be corrected with default values
const DataLoaderContext = createContext(undefined);

const DataLoaderProvider = ({
  children,
  cacheKeyPrefix,
  onError
}) => {
  const [requestQueue, setRequestQueue] = useState({});
  const [cachedData, setCachedDataPrivate] = useState({});
  const reloads = useRef({});
  const computeKey = useCallback(key => `${cacheKeyPrefix ? `${cacheKeyPrefix}-` : ''}${key}`, [cacheKeyPrefix]);
  const setCachedData = useCallback(compute => {
    if (typeof compute === 'function') {
      setCachedDataPrivate(current => compute(current));
    } else {
      setCachedDataPrivate(compute);
    }
  }, []);
  const setReloads = useCallback(compute => {
    if (typeof compute === 'function') {
      reloads.current = compute(reloads.current);
    } else {
      reloads.current = compute;
    }
  }, []);
  const addCachedData = useCallback((key, newData) => {
    if (newData) {
      if (key && typeof key === 'string') {
        setCachedData(actualCachedData => ({ ...actualCachedData,
          [computeKey(key)]: newData
        }));
      } else throw new Error(KEY_IS_NOT_STRING_ERROR);
    }
  }, [setCachedData, computeKey]);
  const addReload = useCallback((key, method) => {
    if (method) {
      if (key && typeof key === 'string') {
        setReloads(actualReloads => ({ ...actualReloads,
          [computeKey(key)]: method
        }));
      } else throw new Error(KEY_IS_NOT_STRING_ERROR);
    }
  }, [setReloads, computeKey]);
  const addRequest = useCallback((key, args) => {
    if (key && typeof key === 'string') {
      const notifyChanges = updatedRequest => {
        setRequestQueue(current => ({ ...current,
          [computeKey(updatedRequest.key)]: updatedRequest
        }));
      };

      const newRequest = new DataLoader({ ...args,
        notify: notifyChanges
      });
      newRequest.addOnSuccessListener(result => {
        if (result !== undefined && result !== null) addCachedData(key, result);
      });
      setRequestQueue(current => ({ ...current,
        [computeKey(key)]: newRequest
      }));
      addReload(key, () => newRequest.load(true));
      return newRequest;
    }

    throw new Error(KEY_IS_NOT_STRING_ERROR);
  }, [computeKey, addCachedData, addReload]);
  const getRequest = useCallback(key => requestQueue[computeKey(key)], [computeKey, requestQueue]);
  const clearReload = useCallback(key => {
    if (key && typeof key === 'string') {
      setReloads(actualReloads => {
        const tmp = actualReloads;
        delete tmp[computeKey(key)];
        return tmp;
      });
    } else throw new Error(KEY_IS_NOT_STRING_ERROR);
  }, [setReloads, computeKey]);
  const clearAllReloads = useCallback(() => {
    setReloads({});
  }, [setReloads]);
  const clearCachedData = useCallback(key => {
    if (key && typeof key === 'string') {
      setCachedData(actualCachedData => {
        const tmp = actualCachedData;
        delete tmp[computeKey(key)];
        return tmp;
      });
    } else throw new Error(KEY_IS_NOT_STRING_ERROR);
  }, [setCachedData, computeKey]);
  const clearAllCachedData = useCallback(() => {
    setCachedData({});
  }, [setCachedData]);
  const reload = useCallback(async key => {
    if (key && typeof key === 'string') {
      var _reloads$current$comp, _reloads$current;

      await ((_reloads$current$comp = (_reloads$current = reloads.current)[computeKey(key)]) === null || _reloads$current$comp === void 0 ? void 0 : _reloads$current$comp.call(_reloads$current));
    } else throw new Error(KEY_IS_NOT_STRING_ERROR);
  }, [computeKey]);
  const reloadAll = useCallback(async () => {
    await Promise.all(Object.values(reloads.current).map(reloadFn => reloadFn()));
  }, []);
  const getCachedData = useCallback(key => {
    if (key) {
      return cachedData[computeKey(key)];
    }

    return cachedData;
  }, [computeKey, cachedData]);
  const getReloads = useCallback(key => {
    if (key) {
      return reloads.current[computeKey(key)];
    }

    return reloads.current;
  }, [computeKey]);
  const value = useMemo(() => ({
    addCachedData,
    addReload,
    addRequest,
    cacheKeyPrefix,
    clearAllCachedData,
    clearAllReloads,
    clearCachedData,
    clearReload,
    getCachedData,
    getReloads,
    getRequest,
    onError,
    reload,
    reloadAll
  }), [addCachedData, addReload, addRequest, cacheKeyPrefix, clearAllCachedData, clearAllReloads, clearCachedData, clearReload, getCachedData, getRequest, getReloads, onError, reload, reloadAll]);
  return /*#__PURE__*/React.createElement(DataLoaderContext.Provider, {
    value: value
  }, children);
};

DataLoaderProvider.propTypes = {
  cacheKeyPrefix: PropTypes.string,
  children: PropTypes.node.isRequired,
  onError: PropTypes.func
};
DataLoaderProvider.defaultProps = {
  cacheKeyPrefix: undefined,
  onError: undefined
};
const useDataLoaderContext = () => useContext(DataLoaderContext);

/**
 * @param {string} key key to save the data fetched in a local cache
 * @param {() => PromiseType} method a method that return a promise
 * @param {useDataLoaderConfig} config hook configuration
 * @returns {useDataLoaderResult} hook result containing data, request state, and method to reload the data
 */
const useDataLoader = (fetchKey, method, {
  enabled = true,
  initialData,
  keepPreviousData = true,
  onError,
  onSuccess,
  pollingInterval,
  maxDataLifetime
} = {}) => {
  const {
    addRequest,
    getCachedData,
    getRequest,
    onError: onErrorProvider
  } = useDataLoaderContext();
  const data = useMemo(() => getCachedData(fetchKey) || initialData, [getCachedData, fetchKey, initialData]);
  const request = useMemo(() => getRequest(fetchKey) ?? addRequest(fetchKey, {
    key: fetchKey,
    maxDataLifetime,
    method,
    pollingInterval
  }), [addRequest, fetchKey, getRequest, method, pollingInterval, maxDataLifetime]);
  useEffect(() => {
    if (enabled && request.status === StatusEnum.IDLE) {
      // eslint-disable-next-line no-void
      void request.load();
    }
  }, [request, enabled]);
  useEffect(() => {
    if (request.method !== method) {
      request.method = method;
    }

    request.addOnErrorListener(onError ?? onErrorProvider);
    request.addOnSuccessListener(onSuccess);
    return () => {
      request.removeOnErrorListener(onError ?? onErrorProvider);
      request.removeOnSuccessListener(onSuccess);
    };
  }, [onSuccess, onError, onErrorProvider, method, request]);
  const cancelMethodRef = useRef(request === null || request === void 0 ? void 0 : request.cancel);
  const isMountedRef = useRef(false);
  const isFetchingRef = useRef(false);
  const previousDataRef = useRef();
  const isLoading = useMemo(() => request.status === StatusEnum.LOADING, [request.status]);
  const isIdle = useMemo(() => request.status === StatusEnum.IDLE, [request.status]);
  const isSuccess = useMemo(() => request.status === StatusEnum.SUCCESS, [request.status]);
  const isError = useMemo(() => request.status === StatusEnum.ERROR, [request.status]);
  const isPolling = useMemo(() => !!(enabled && pollingInterval && (isSuccess || isLoading)), [isSuccess, isLoading, enabled, pollingInterval]);
  useEffect(() => {
    isFetchingRef.current = isLoading || isPolling;
  }, [isLoading, isPolling]);
  useLayoutEffect(() => {
    cancelMethodRef.current = request === null || request === void 0 ? void 0 : request.cancel;
  }, [request === null || request === void 0 ? void 0 : request.cancel]);
  useEffect(() => {
    isMountedRef.current = true;
    return () => {
      isMountedRef.current = false;

      if (isFetchingRef.current && cancelMethodRef.current) {
        cancelMethodRef.current();
      }
    };
  }, []);
  useEffect(() => () => {
    if (data !== previousDataRef.current) {
      previousDataRef.current = data;
    }
  }, [data, keepPreviousData]);
  return {
    data: getCachedData(fetchKey) || initialData,
    error: request === null || request === void 0 ? void 0 : request.error,
    isError,
    isIdle,
    isLoading,
    isPolling,
    isSuccess,
    previousData: previousDataRef.current,
    reload: () => request.load(true)
  };
};

export { DataLoaderProvider, useDataLoader, useDataLoaderContext };
